/* 
 * PVM++ - A C++ library for easy use of PVM
 * Copyright (C) 1997-2001 Sebastian Wilhelmi; University of Karlsruhe
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "pvm++/private.hh"
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#define MICROSEC_PER_SEC 1000000

namespace Pvm
{

  namespace Internal
  {
  
    void
    CalcEndTime (unsigned long int Time,
		 long int &Sec, long int &uSec)
    {
      struct timeval TimeNow;
      gettimeofday (&TimeNow, NULL);
      if (Time % MICROSEC_PER_SEC + TimeNow.tv_usec < MICROSEC_PER_SEC)
	{
	  Sec = TimeNow.tv_sec + Time / MICROSEC_PER_SEC;
	  uSec = TimeNow.tv_usec + Time % MICROSEC_PER_SEC;
	}
      else
	{
	  Sec = TimeNow.tv_sec + Time / MICROSEC_PER_SEC + 1;
	  uSec = TimeNow.tv_usec + Time % MICROSEC_PER_SEC - MICROSEC_PER_SEC;
	}
    }

    unsigned long int
    CalcRestTime (long int Sec, long int uSec)
    {
      struct timeval TimeNow;
      gettimeofday (&TimeNow, NULL);
      // FIXME: if timeval.tv_sec and timeval.tv_usec are unsigned, a
      // warning message appeares
      if (Sec < TimeNow.tv_sec ||
	  Sec == TimeNow.tv_sec && uSec <= TimeNow.tv_usec)
	return 0;

      return (unsigned long) (Sec - TimeNow.tv_sec) * MICROSEC_PER_SEC +
	(unsigned long) uSec - TimeNow.tv_usec;
    }

    void
    CalcRestTime (long int Sec, long int uSec,
		  long int &dSec, long int &duSec)
    {
      struct timeval TimeNow;
      gettimeofday (&TimeNow, NULL);
      // FIXME: if timeval.tv_sec and timeval.tv_usec are unsigned, a
      // warning message appeares
      if (Sec < TimeNow.tv_sec ||
	  Sec == TimeNow.tv_sec && uSec <= TimeNow.tv_usec)
	{
	  dSec = 0;
	  duSec = 0;
	}
      else if (uSec >= TimeNow.tv_usec)
	{
	  dSec = Sec - TimeNow.tv_sec;
	  duSec = uSec - TimeNow.tv_usec;
	}
      else
	{
	  dSec = Sec - 1 - TimeNow.tv_sec;
	  duSec = uSec + MICROSEC_PER_SEC - TimeNow.tv_usec;
	}
    }

    void
    GetTasks (int What, TaskSet &Result)
    {
      Result.erase (Result.begin (), Result.end ());
      pvmtaskinfo *Info;
      int NumOfTasks;
      THROW (pvm_tasks (What, &NumOfTasks, &Info));
      for (int i = 0; i < NumOfTasks; ++i)
	{
	  TaskTable->Add (Info[i].ti_tid, &Info[i], false);
	  Result.insert (Info[i].ti_tid);
	}
    }

    QueueIterator 
    GetBuffer (int Buffer)
    {
      THROW (Buffer);
      StructId StructId;
      int TaskId;
      int Dummy;
      THROW (pvm_bufinfo (Buffer, &Dummy, (int *) &StructId, &TaskId));
      if (StructId == HostDeleteInfo)
	{
	  int HostId;
	  pvm_upkint (&HostId, 1, 1);
	  HostTable->Delete (HostId);
	  return ReceivedQueue->end ();
	}
      if (StructId == HostAddInfo)
	{
	  HostTable->Refresh ();
	  return ReceivedQueue->end ();
	}
      THROW (pvm_setrbuf (0));
      StructId -= StructIdOffset;
      if (!HandlerTable->ExecuteHandler (Buffer, StructId, TaskId))
	{
	  Received New;
	  New.BufferId = Buffer;
	  New.StructId = StructId;
	  New.Task = TaskId;
	  ReceivedQueue->push_back (New);
	  return --ReceivedQueue->end ();
	}
      return ReceivedQueue->end ();
    }

    QueueIterator
    ReceiveIt ()
    {
      QueueIterator Result;
      do
	{
	  Result = GetBuffer (pvm_recv (-1, -1));
	}
      while (Result == ReceivedQueue->end ());
      return Result;
    }

    QueueIterator
    ReceiveIt (long int EndSec, long int EnduSec)
    {
      QueueIterator Result;
      do
	{
	  timeval TimeOut;
	  CalcRestTime (EndSec, EnduSec, (long int &) TimeOut.tv_sec,
			(long int &) TimeOut.tv_usec);
	  int Buffer = pvm_trecv (-1, -1, &TimeOut);
	  THROW (Buffer);

	  if (Buffer == 0)
	    return ReceivedQueue->end ();

	  Result = GetBuffer (Buffer);
	}
      while (Result == ReceivedQueue->end ());
      return Result;
    }

    QueueIterator 
    ReceiveItNoBlock ()
    {
      QueueIterator Result;
      do
	{
	  int Buffer = pvm_nrecv (-1, -1);
	  THROW (Buffer);

	  if (Buffer == 0)
	    return ReceivedQueue->end ();
	
	  Result = GetBuffer (Buffer);
	}
      while (Result == ReceivedQueue->end ());
      return Result;
    }

    bool
    Search (const TaskSet &TaskSet,
	    const StructSet &StructSet,
	    QueueIterator &QueuePos)
    {
      QueueIterator Last = ReceivedQueue->end ();
      while (QueuePos != Last)
	{
	  if (TaskSet.count ((*QueuePos).Task) &&
	      StructSet.count ((*QueuePos).StructId))
	    return true;

	  ++QueuePos;
	}
      return false;
    }

    bool
    Search (const StructSet &StructSet,
	    QueueIterator &QueuePos)
    {
      QueueIterator Last = ReceivedQueue->end ();
      while (QueuePos != Last)
	{
	  if (StructSet.count ((*QueuePos).StructId))
	    return true;

	  ++QueuePos;
	}
      return false;
    }

    HostTableType *HostTable;
    TaskTableType *TaskTable;
    HandlerTableType *HandlerTable;
    QueueType *ReceivedQueue;
  
  } // namespace Internal
  
  using namespace Internal;

  void
  AccessPrivate::UnPack (const QueueIterator &QueuePos, StructSet &What,
			 Task &From, StructId &StructId)
  {
    StructId = (*QueuePos).StructId;
    From = (*QueuePos).Task;
    UnPack (QueuePos, *(*What.Map.find (StructId)).second);
  }

  void
  AccessPrivate::UnPack (const QueueIterator &QueuePos, Struct &What)
  {
    int Buffer = (*QueuePos).BufferId;
    ReceivedQueue->erase (QueuePos);
    THROW (pvm_setrbuf (Buffer));
    What.UnPack ();
    THROW (pvm_freebuf (Buffer));
  }

  QueueIterator 
  AccessPrivate::ReceiveSelect (StructSet& Set, long int Sec, long int uSec)
  {
    if (Set.UseSelectUnknown)
      Set.UseSelect = !(Set.IntReadFDs.empty () && Set.IntWriteFDs.empty ()
			&& Set.IntExceptFDs.empty ());
    
    if (Set.UseSelect)
      {
	if (Select (Set, Sec, uSec))
	  return ReceiveItNoBlock ();
	else
	  return ReceivedQueue->end ();
      }
    else if (Sec == 0 && uSec == 0)
      return ReceiveIt ();
    else
      return ReceiveIt (Sec, uSec);
  }
  
  bool
  AccessPrivate::Select (StructSet &StructSet, long int EndSec,
			 long int EnduSec)
  {
    fd_set ReadFDs, WriteFDs, ExceptFDs;
    FD_ZERO (&ReadFDs);
    FD_ZERO (&WriteFDs);
    FD_ZERO (&ExceptFDs);
    int *PvmFDs;
    int EndFD = 0;
    int NumPvmFD = pvm_getfds (&PvmFDs);
    THROW (NumPvmFD);
    for (int i = 0; i < NumPvmFD; ++i)
      {
	FD_SET (PvmFDs[i], &ReadFDs);
	EndFD = std::max (EndFD, PvmFDs[i] + 1);
      }
    StructSet::FDSet::iterator Current = StructSet.IntReadFDs.begin ();
    for (; Current != StructSet.IntReadFDs.end (); ++Current)
      {
	FD_SET (*Current, &ReadFDs);
	EndFD = std::max (EndFD, *Current + 1);
      }
    Current = StructSet.IntWriteFDs.begin ();
    for (; Current != StructSet.IntWriteFDs.end (); ++Current)
      {
	FD_SET (*Current, &WriteFDs);
	EndFD = std::max (EndFD, *Current + 1);
      }
    Current = StructSet.IntExceptFDs.begin ();
    for (; Current != StructSet.IntExceptFDs.end (); ++Current)
      {
	FD_SET (*Current, &ExceptFDs);
	EndFD = std::max (EndFD, *Current + 1);
      }
    timeval *SelectTime = NULL;
    if (EndSec != 0 || EnduSec != 0)
      {
	timeval TimeOut;
	CalcRestTime (EndSec, EnduSec, (long int &) TimeOut.tv_sec,
		      (long int &) TimeOut.tv_usec);
	SelectTime = &TimeOut;
      }
    int NumOfReadyFDs = select (EndFD, (FD_SET_TYPE *) &ReadFDs,
				(FD_SET_TYPE *) &WriteFDs,
				(FD_SET_TYPE *) &ExceptFDs,
				SelectTime);
    bool PvmReady = false;
    for (int i = 0; i < NumPvmFD && NumOfReadyFDs != 0; ++i)
      {
	if (FD_ISSET (PvmFDs[i], &ReadFDs))
	  {
	    NumOfReadyFDs--;
	    FD_CLR (PvmFDs[i], &ReadFDs);
	    PvmReady = true;
	  }
      }
    StructSet.IntReadyReadFDs.erase (StructSet.IntReadyReadFDs.begin (),
				     StructSet.IntReadyReadFDs.end ());
    StructSet.IntReadyWriteFDs.erase (StructSet.IntReadyWriteFDs.begin (),
				      StructSet.IntReadyWriteFDs.end ());
    StructSet.IntReadyExceptFDs.erase (StructSet.IntReadyExceptFDs.begin (),
				       StructSet.IntReadyExceptFDs.end ());
    StructSet.IntFDsReady = NumOfReadyFDs > 0;
    if (StructSet.IntFDsReady)
      {
	for (int i = 0; i < EndFD && NumOfReadyFDs != 0; ++i)
	  {
	    if (FD_ISSET (i, &ReadFDs))
	      {
		NumOfReadyFDs--;
		StructSet.IntReadyReadFDs.insert (i);
	      }
	    if (FD_ISSET (i, &WriteFDs))
	      {
		NumOfReadyFDs--;
		StructSet.IntReadyWriteFDs.insert (i);
	      }
	    if (FD_ISSET (i, &ExceptFDs))
	      {
		NumOfReadyFDs--;
		StructSet.IntReadyExceptFDs.insert (i);
	      }
	  }
      }
    return PvmReady;
  }
  
} // namespace Pvm
